"""Fund Holdings Standard Model."""

from typing import Optional

from openbb_core.provider.abstract.data import Data
from openbb_core.provider.abstract.query_params import QueryParams
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from pydantic import Field, field_validator


class FundHoldingsQueryParams(QueryParams):
    """Fund Holdings Query."""

    symbol: str = Field(description=QUERY_DESCRIPTIONS.get("symbol", "") + " (Fund)")

    @field_validator("symbol")
    @classmethod
    def to_upper(cls, v: str) -> str:
        """Convert field to uppercase."""
        return v.upper()


class FundHoldingsData(Data):
    """Fund Holdings Data."""

    symbol: Optional[str] = Field(
        default=None, description=DATA_DESCRIPTIONS.get("symbol", "") + " (Fund)"
    )
    name: Optional[str] = Field(
        default=None,
        description="Name of the Fund holding.",
    )



import logging
import pandas as pd
import akshare as ak
from typing import Optional, Literal
from openbb_akshare import project_name
from mysharelib.tools import setup_logger, normalize_symbol

setup_logger(project_name)
logger = logging.getLogger(__name__)

def fetch_key_metrics(
        symbol: str, 
        period: Literal["annual", "quarter"] = "quarter",
        use_cache: bool = True,
        api_key: Optional[str] = None,
        ) -> pd.DataFrame:
    """
    Fetches key financial metrics for a specific equity symbol.

    Args:
        symbol (str): The stock symbol to fetch metrics for.
                      such as "601127.SH".

    Returns:
        pd.DataFrame: A DataFrame containing the key metrics.
    """
    from mysharelib.blob_cache import BlobCache

    logger.debug(f"Fetching key metrics for symbol: {symbol}, period: {period}, use_cache: {use_cache}")
    symbol_b, _, market = normalize_symbol(symbol)
    if market not in ["SH", "SZ", "BJ", "HK"]:
        logger.warning("AKShare key metrics only support A shares.")
        return pd.DataFrame()
    cache = BlobCache(table_name="key_metrics", project=project_name)
    try:
        data = cache.load_cached_data(symbol_b, period, use_cache, _get_key_metrics, api_key=api_key)
    except NotImplementedError:
        logger.warning("Cached data contained pandas extension dtypes that could not be unpickled; refreshing cache.")
        # Try again without using cache so we regenerate a fresh, safe cache entry
        data = cache.load_cached_data(symbol_b, period, False, _get_key_metrics, api_key=api_key)

    if data is None:
        return pd.DataFrame()

    # Ensure loaded DataFrame has no pandas extension dtypes (e.g., StringDtype)
    if isinstance(data, pd.DataFrame):
        for col in data.columns:
            if pd.api.types.is_extension_array_dtype(data[col].dtype):
                data[col] = data[col].astype(object)
        data.index = data.index.astype(object)

        # If cached data is missing required fields (e.g., '证券代码'), regenerate cache
        if "证券代码" not in data.index:
            logger.warning("Cached key metrics missing '证券代码'; refreshing cache.")
            data = cache.load_cached_data(symbol_b, period, False, _get_key_metrics, api_key=api_key)
            if isinstance(data, pd.DataFrame):
                for col in data.columns:
                    if pd.api.types.is_extension_array_dtype(data[col].dtype):
                        data[col] = data[col].astype(object)
                data.index = data.index.astype(object)

    return data

def _get_key_metrics(
    symbol: str,
    period: str = "quarter",
    api_key: Optional[str] = None
) -> pd.DataFrame:
    import akshare as ak
    from mysharelib.em.get_a_info_em import get_a_info_em
    from mysharelib.em.get_hk_info_em import get_hk_info_em

    symbol_b, symbol_f, market = normalize_symbol(symbol)
    df_base = pd.DataFrame()
    if market == "HK":
        df_base, _ = get_hk_info_em(symbol_f)
    else:
        df_base, _ = get_a_info_em(symbol_f)

    # Convert any existing StringDtype or extension array columns to object type immediately
    for col in df_base.columns:
        if pd.api.types.is_extension_array_dtype(df_base[col].dtype) or pd.api.types.is_string_dtype(df_base[col]):
            df_base[col] = df_base[col].astype(object)

    if api_key:
        try:
            ak.stock.cons.xq_a_token = api_key
            stock_individual_spot_xq_df = ak.stock_individual_spot_xq(symbol=f"{market}{symbol_b}")

            def get_metric(df, metric_name):
                result = df.loc[df["item"] == metric_name, "value"]
                if not result.empty:
                    return result.iloc[0]
                return None

            # Add new metrics and ensure they're proper types
            metrics_to_add = {
                "市盈率(动)": get_metric(stock_individual_spot_xq_df, "市盈率(动)"),
                "市盈率(TTM)": get_metric(stock_individual_spot_xq_df, "市盈率(TTM)"),
                "市盈率(静)": get_metric(stock_individual_spot_xq_df, "市盈率(静)"),
                "市净率": get_metric(stock_individual_spot_xq_df, "市净率"),
                "流通值": get_metric(stock_individual_spot_xq_df, "流通值"),
                "52周最低": get_metric(stock_individual_spot_xq_df, "52周最低"),
                "52周最高": get_metric(stock_individual_spot_xq_df, "52周最高"),
                "股息(TTM)": get_metric(stock_individual_spot_xq_df, "股息(TTM)"),
                "股息率(TTM)": get_metric(stock_individual_spot_xq_df, "股息率(TTM)"),
                "发行日期": get_metric(stock_individual_spot_xq_df, "发行日期")
            }
            
            # Add metrics to df_base ensuring proper data types
            for metric_name, value in metrics_to_add.items():
                df_base.loc[metric_name] = str(value) if value is not None else None
        except NotImplementedError:
            logger.warning("akshare returned an unexpected format; skipping extra spot metrics.")
        except Exception:
            logger.warning("Failed to fetch additional spot metrics from akshare; proceeding without them.")

    # Ensure symbol field is present for downstream data model validation
    if "证券代码" not in df_base.index:
        df_base.loc["证券代码"] = symbol

    # Final comprehensive conversion to ensure no pandas extension dtypes (e.g., StringDtype) remain
    df_converted = df_base.copy()
    for col in df_converted.columns:
        if pd.api.types.is_extension_array_dtype(df_converted[col].dtype):
            df_converted[col] = df_converted[col].astype(object)
        else:
            df_converted[col] = df_converted[col].astype(object)
    # Ensure index values are plain Python objects (avoid extension dtypes)
    df_converted.index = df_converted.index.astype(object)
    
    return df_converted


import logging
import pandas as pd
from mysharelib.tools import setup_logger, get_exchange
from mysharelib.table_cache import TableCache
from openbb_akshare import project_name

setup_logger(project_name)
logger = logging.getLogger(__name__)

TABLE_SCHEMA = {
    "symbol": "TEXT",               # Symbol (e.g. 600000)
    "name": "TEXT",                 # Short name of the asset
    "exchange": "TEXT"              # Exchange code (e.g. SSE/SZSE)
}

def get_symbols_df() -> pd.DataFrame:
    import akshare as ak

    stock_info_a_code_name_df = ak.stock_info_a_code_name()
    stock_info_a_code_name_df.rename(columns={"code": "symbol"}, inplace=True)
    stock_info_a_code_name_df['exchange'] = [
        get_exchange(code)
        for code in stock_info_a_code_name_df['symbol']
        ]
    # Get symbols for HK market
    hk_stock_list_df = ak.stock_hk_spot_em()
    stock_info_hk_code_name_df = hk_stock_list_df[["代码","名称"]].copy()
    stock_info_hk_code_name_df.rename(columns={"代码": "symbol", "名称": "name"}, inplace=True)
    stock_info_hk_code_name_df["exchange"] = "HKEX"
    
    return pd.concat([stock_info_a_code_name_df, stock_info_hk_code_name_df], ignore_index=True) 

def get_symbols(use_cache: bool = True, api_key : str = "") -> pd.DataFrame:
    cache = TableCache(TABLE_SCHEMA, project=project_name, table_name="symbols", primary_key="symbol")
    if use_cache:
        data = cache.read_dataframe()
        if not data.empty:
            logger.info(f"Loading symbols from {project_name} cache...")
            return data

    logger.info(f"Generating symbols for {project_name} ...")
    data = get_symbols_df()
    cache.write_dataframe(data)
    return data


"""AKShare Equity Quote Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional
from warnings import warn

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_quote import (
    EquityQuoteData,
    EquityQuoteQueryParams,
)
from pydantic import Field
import logging
from mysharelib.tools import setup_logger, normalize_symbol
from openbb_akshare import project_name

setup_logger(project_name)
logger = logging.getLogger(__name__)

class AKShareEquityQuoteQueryParams(EquityQuoteQueryParams):
    """AKShare Equity Quote Query."""

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}

    use_cache: bool = Field(
        default=True,
        description="Whether to use a cached request. The quote is cached for one hour.",
    )


class AKShareEquityQuoteData(EquityQuoteData):
    """AKShare Equity Quote Data."""

    __alias_dict__ = {
        "symbol": "代码",
        "name": "名称",
        "last_price": "最新价",
        "open": "今开",
        "high": "最高",
        "low": "最低",
        "prev_close": "昨收",
        "volume": "成交量",
        "turnover": "成交额",
        "change": "涨跌额",
        "change_percent": "涨幅",
        "amplitude": "振幅",
        "pe_ratio": "市盈率",
        "pb_ratio": "市净率",
        "52_week_high": "52周最高",
        "52_week_low": "52周最低",
        "float_shares": "流通股",
        "limit_down": "跌停",
        "float_market_value": "流通值",
        "lot_size": "最小交易单位",
        "price_change": "涨跌",
        "eps": "每股收益",
        "turnover_rate": "周转率",  # alternate wording for 换手率/周转率
        "exchange": "交易所",
        "pe_ratio_dynamic": "市盈率(动)",
        "funds_share_ratio": "基金份额/总股本",
        "goodwill_in_net_assets": "净资产中的商誉",
        "average_price": "均价",
        "current_price": "现价",
        "ytd_change": "今年以来涨幅",
        "issue_date": "发行日期",
        "net_asset_to_market_cap": "资产净值/总市值",
        "dividend_ttm": "股息(TTM)",
        "dividend_yield_ttm": "股息率(TTM)",
        "currency": "货币",
        "book_value_per_share": "每股净资产",
        "pe_ratio_static": "市盈率(静)",
        "limit_up": "涨停",
        "pe_ratio_ttm": "市盈率(TTM)",
        "time": "时间",
        "ps_ratio": "市销率",
    }


class AKShareEquityQuoteFetcher(
    Fetcher[AKShareEquityQuoteQueryParams, List[AKShareEquityQuoteData]]
):
    """AKShare Equity Quote Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> AKShareEquityQuoteQueryParams:
        """Transform the query."""
        return AKShareEquityQuoteQueryParams(**params)

    @staticmethod
    def extract_data(
        query: AKShareEquityQuoteQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from AKShare."""
        # pylint: disable=import-outside-toplevel
        import pandas as pd

        api_key = credentials.get("akshare_api_key", "") if credentials else ""

        symbols = query.symbol.split(",")
        all_data = []

        if not api_key:
            logger.error("AKShare requires an API key.")
            return all_data

        def get_one(symbol, api_key: str, use_cache) -> pd.DataFrame:
            """Get the data for one ticker symbol."""
            quote = pd.DataFrame()
            symbol_b, symbol_f, market = normalize_symbol(symbol)

            if market == "HK":
                symbol_xq = symbol_b
            else:
                symbol_xq = f"{market}{symbol_b}"

            logger.info(f"Fetching data for symbol: {symbol_xq}")

            import akshare as ak

            ak.stock.cons.xq_a_token = api_key
            stock_individual_spot_xq_df = ak.stock_individual_spot_xq(symbol=symbol_xq)

            if stock_individual_spot_xq_df.empty:
                return pd.DataFrame([{"symbol": symbol, "error": "Symbol not found"}])
            else:
                data = stock_individual_spot_xq_df.set_index("item")
                #logger.info(f"Fetched data for symbol: {data}")
                return data.T

        for symbol in symbols:
            try:        
                data = get_one(symbol, api_key=api_key, use_cache=query.use_cache)
                all_data.append(data.to_dict(orient="records")[0])
                
            except Exception as e:
                print(f"Error fetching data for symbol {symbol}: {e}")
                continue

        return all_data

    @staticmethod
    def transform_data(
        query: AKShareEquityQuoteQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[AKShareEquityQuoteData]:
        """Transform the data."""
        return [AKShareEquityQuoteData.model_validate(d) for d in data]



"""Akshare Fund Holdings Model."""

# pylint: disable=unused-argument

from datetime import (
    date as dateType,
    datetime,
)
import re
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from openbb_akshare.standard_models.fund_holdings import (
    FundHoldingsData,
    FundHoldingsQueryParams,
)
from openbb_akshare.utils.helpers import ak_fund_portfolio_hold_em
from openbb_core.provider.abstract.data import ForceInt
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class AkshareFundHoldingsQueryParams(FundHoldingsQueryParams):
    """FMP Fund Holdings Query.

    Source: https://fundf10.eastmoney.com/ccmx_000001.html
    """

    date: Optional[Union[str, dateType]] = Field(
        description=QUERY_DESCRIPTIONS.get("date", "")
        + " Entering a date will attempt to return the NPORT-P filing for the entered date."
        + " For Provider Akshare, only the year of date functions."
        + " Use the holdings_date command/endpoint to find available filing dates for the Fund.",
        default="2024-01-01",
    )
    use_cache: Optional[bool] = Field(
        default=True,
        description="Whether or not to use cache. If True, cache will store for two days.",
    )

    @field_validator("date", mode="before")
    @classmethod
    def validate_date_format(cls, v):
        """Accept YYYY-MM-DD strings or date objects; reject other formats with a clear message."""
        if v is None:
            return v
        if isinstance(v, dateType):
            return v
        s = str(v).strip()
        if not re.fullmatch(r"\d{4}-\d{2}-\d{2}", s):
            raise ValueError("Invalid 'date' format. Expected YYYY-MM-DD (e.g. 2024-01-01).")
        return s

    @field_validator("symbol", mode="before", check_fields=False)
    @classmethod
    def normalize_symbol(cls, v):
        """Accept Yahoo-style fund symbols and normalize to the 6-digit fund code.

        Examples accepted:
        - 000001.OF
        - 000001
        - OF000001
        """
        if v is None:
            raise ValueError("'symbol' is required. Example: 000001.OF")
        raw = str(v).strip().upper()
        if not raw:
            raise ValueError("'symbol' cannot be empty. Example: 000001.OF")
        raw = raw.split(",")[0].strip()

        # Normalize suffix/prefix
        raw = raw.replace(".OF", "")
        raw = raw.replace(".SS", "")
        raw = raw.replace(".SH", "")
        raw = raw.replace(".SZ", "")
        raw = raw.replace(".BJ", "")
        if raw.startswith(("OF", "SH", "SZ", "BJ")):
            raw = raw[2:]

        if not re.fullmatch(r"\d{6}", raw):
            raise ValueError(
                "Invalid 'symbol' format for fund holdings. Use Yahoo-style like '000001.OF' (or raw 6 digits)."
            )
        return raw


class AkshareFundHoldingsData(FundHoldingsData):
    """FMP Fund Holdings Data."""

    __alias_dict__ = {
        "name": "股票名称",
        "balance": "持股数",
        "value": "持仓市值",
        "weight": "占净值比例",
        "acceptance_datetime": "季度",
    }

    balance: Optional[ForceInt] = Field(
        description="The balance of the holding, in shares or units.", default=None
    )
    value: Optional[float] = Field(
        description="The value of the holding, in dollars.",
        default=None,
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    weight: Optional[float] = Field(
        description="The weight of the holding, as a normalized percent.",
        default=None,
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    acceptance_datetime: Optional[str] = Field(
        description="The acceptance datetime of the filing.",
        default=None,
    )

    @field_validator("weight", mode="before", check_fields=False)
    @classmethod
    def normalize_percent(cls, v):
        """Normalize percent values."""
        return float(v) if v else None

    @field_validator("balance", "name", "symbol", "value")
    @classmethod
    def replace_empty(cls, v):
        """Replace empty strings and 0s with None."""
        if isinstance(v, str):
            return v if v not in ("", "0", "-") else None
        if isinstance(v, (float, int)):
            return v if v and v not in (0.0, 0) else None
        return v if v else None


class AkshareFundHoldingsFetcher(
    Fetcher[
        AkshareFundHoldingsQueryParams,
        List[AkshareFundHoldingsData],
    ]
):
    """Transform the query, extract and transform the data from the FMP endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> AkshareFundHoldingsQueryParams:
        """Transform the query."""
        return AkshareFundHoldingsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: AkshareFundHoldingsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from Akshare."""
        if isinstance(query.date, str):
            try:
                query.date = datetime.strptime(query.date, "%Y-%m-%d")
            except ValueError as e:
                raise ValueError(
                    f"Invalid 'date' format: '{query.date}'. Expected YYYY-MM-DD (e.g. 2024-01-01)."
                ) from e

        try:
            fund_portfolio_hold_em_df = ak_fund_portfolio_hold_em(
                symbol=query.symbol,
                year=str(query.date.year),
                db_path="etf_fund_holdings",
                use_cache=query.use_cache,
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to fetch fund holdings for symbol '{query.symbol}'. If you passed a Yahoo symbol like 000001.OF, it should be supported. Underlying error: {e}"
            ) from e

        if fund_portfolio_hold_em_df is None or getattr(fund_portfolio_hold_em_df, "empty", True):
            raise EmptyDataError(f"No fund holdings data found for symbol '{query.symbol}' and date '{query.date}'.")

        fund_portfolio_hold_em_df.drop(["股票代码", "序号", "code"], axis=1, inplace=True)

        return fund_portfolio_hold_em_df.to_dict(orient="records")

    @staticmethod
    def transform_data(
        query: AkshareFundHoldingsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[AkshareFundHoldingsData]:
        """Return the transformed data."""
        # Limited to one alias per field, so we need to do these here.
        for i in data:
            if "SH" in i["symbol"]:
                i["symbol"] = i["symbol"].replace("SH", "") + ".SS"
            elif "SZ" in i["symbol"]:
                i["symbol"] = i["symbol"].replace("SZ", "") + ".SZ"
            elif "OF" in i["symbol"]:
                i["symbol"] = i["symbol"].replace("OF", "") + ".OF"
            elif "BJ" in i["symbol"]:
                i["symbol"] = i["symbol"].replace("BJ", "") + ".BJ"
        return [AkshareFundHoldingsData.model_validate(d) for d in data]
